from celery import shared_task, chain

from django.db import IntegrityError

from eveuniverse.core.esitools import is_esi_online
from eveuniverse.tasks import update_unresolved_eve_entities

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from . import __title__, APP_NAME
from .app_settings import (
    KILLTRACKER_MAX_KILLMAILS_PER_RUN,
    KILLTRACKER_STORING_KILLMAILS_ENABLED,
    KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS,
    KILLTRACKER_TASKS_TIMEOUT,
    KILLTRACKER_DISCORD_SEND_DELAY,
    KILLTRACKER_GENERATE_MESSAGE_MAX_RETRIES,
    KILLTRACKER_GENERATE_MESSAGE_RETRY_COUNTDOWN,
    KILLTRACKER_TASK_OBJECTS_CACHE_TIMEOUT,
    KILLTRACKER_TASK_MINIMUM_RETRY_DELAY,
)
from .core.killmails import Killmail
from .exceptions import WebhookTooManyRequests
from .models import (
    EveKillmail,
    Tracker,
    Webhook,
)
from app_utils.caching import cached_queryset
from app_utils.logging import LoggerAddTag


logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@shared_task(
    bind=True,
    base=QueueOnce,
    timeout=KILLTRACKER_TASKS_TIMEOUT,
    retry_backoff=False,
    max_retries=None,
)
def run_killtracker(self) -> None:
    """Main task for running the Killtracker.
    Will fetch new killmails from ZKB and start running trackers for them
    """
    if not is_esi_online():
        logger.warning("ESI is currently offline. Aborting")
        return

    if self.request.retries == 0:
        logger.info("Killtracker run started...")
        qs = cached_queryset(
            Webhook.objects.filter(is_enabled=True),
            key=f"{APP_NAME}_enabled_webhooks",
            timeout=KILLTRACKER_TASK_OBJECTS_CACHE_TIMEOUT,
        )
        for webhook in qs:
            webhook.reset_failed_messages()

    killmail = Killmail.create_from_zkb_redisq()
    if killmail:
        killmail_json = killmail.asjson()
        qs = cached_queryset(
            Tracker.objects.filter(is_enabled=True),
            key=f"{APP_NAME}_enabled_trackers",
            timeout=KILLTRACKER_TASK_OBJECTS_CACHE_TIMEOUT,
        )
        for tracker in qs:
            run_tracker.delay(
                tracker_pk=tracker.pk,
                killmail_json=killmail_json,
            )

        if KILLTRACKER_STORING_KILLMAILS_ENABLED:
            chain(
                store_killmail.si(killmail_json=killmail_json),
                update_unresolved_eve_entities.si(),
            ).delay()

    total_killmails = self.request.retries + (1 if killmail else 0)
    if killmail and total_killmails < KILLTRACKER_MAX_KILLMAILS_PER_RUN:
        self.retry(countdown=KILLTRACKER_TASK_MINIMUM_RETRY_DELAY)
    else:
        if (
            KILLTRACKER_STORING_KILLMAILS_ENABLED
            and KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS > 0
        ):
            delete_stale_killmails.delay()

        logger.info(
            "Killtracker run completed. %d killmails received from ZKB", total_killmails
        )


@shared_task(timeout=KILLTRACKER_TASKS_TIMEOUT)
def run_tracker(
    tracker_pk: int, killmail_json: str, ignore_max_age: bool = False
) -> None:
    """run tracker for given killmail and trigger sending if needed"""

    tracker = Tracker.objects.get_cached(
        pk=tracker_pk,
        select_related="webhook",
        timeout=KILLTRACKER_TASK_OBJECTS_CACHE_TIMEOUT,
    )
    logger.info("%s: Started running tracker", tracker)
    killmail = Killmail.from_json(killmail_json)
    killmail_new = tracker.process_killmail(
        killmail=killmail, ignore_max_age=ignore_max_age
    )
    if killmail_new:
        generate_killmail_message.delay(
            tracker_pk=tracker_pk, killmail_json=killmail_new.asjson()
        )
    elif tracker.webhook.main_queue.size():
        send_messages_to_webhook.delay(webhook_pk=tracker.webhook.pk)


@shared_task(bind=True, timeout=KILLTRACKER_TASKS_TIMEOUT)
def generate_killmail_message(self, tracker_pk: int, killmail_json: str) -> None:
    """generate and enqueue message from given killmail and start sending"""

    tracker = Tracker.objects.get_cached(
        pk=tracker_pk,
        select_related="webhook",
        timeout=KILLTRACKER_TASK_OBJECTS_CACHE_TIMEOUT,
    )
    killmail_new = Killmail.from_json(killmail_json)
    logger.info("%s: Generating message from killmail %s", tracker, killmail_new.id)
    try:
        tracker.generate_killmail_message(killmail_new)
    except Exception as ex:
        will_retry = self.request.retries < KILLTRACKER_GENERATE_MESSAGE_MAX_RETRIES
        logger.warning(
            "%s: Failed to generate killmail %s.%s",
            tracker,
            killmail_new.id,
            " Will retry." if will_retry else "",
            exc_info=True,
        )
        self.retry(
            max_retries=KILLTRACKER_GENERATE_MESSAGE_MAX_RETRIES,
            countdown=KILLTRACKER_GENERATE_MESSAGE_RETRY_COUNTDOWN,
            exc=ex,
        )
    else:
        send_messages_to_webhook.delay(webhook_pk=tracker.webhook.pk)


@shared_task(timeout=KILLTRACKER_TASKS_TIMEOUT)
def store_killmail(killmail_json: str) -> None:
    """stores killmail as EveKillmail object"""
    killmail = Killmail.from_json(killmail_json)
    try:
        EveKillmail.objects.create_from_killmail(killmail, resolve_ids=False)
    except IntegrityError:
        logger.warning(
            "%s: Failed to store killmail, because it already exists", killmail.id
        )
    else:
        logger.info("%s: Stored killmail", killmail.id)


@shared_task(timeout=KILLTRACKER_TASKS_TIMEOUT)
def delete_stale_killmails() -> None:
    """deleted all EveKillmail objects that are considered stale"""
    _, details = EveKillmail.objects.delete_stale()
    if details:
        logger.info("Deleted %d stale killmails", details["killtracker.EveKillmail"])


@shared_task(
    bind=True,
    base=QueueOnce,  # celery_once locks stay intact during retries
    timeout=KILLTRACKER_TASKS_TIMEOUT,
    retry_backoff=False,
    max_retries=None,
)
def send_messages_to_webhook(self, webhook_pk: int) -> None:
    """send all queued messages to given Webhook"""

    webhook = Webhook.objects.get_cached(
        pk=webhook_pk,
        timeout=KILLTRACKER_TASK_OBJECTS_CACHE_TIMEOUT,
    )
    if not webhook.is_enabled:
        logger.info("%s: Webhook is disabled - aborting", webhook)
        return

    message = webhook.main_queue.dequeue()
    if message:
        logger.info("%s: Sending message to webhook", webhook)
        try:
            response = webhook.send_message_to_webhook(message)
        except WebhookTooManyRequests as ex:
            webhook.main_queue.enqueue(message)
            logger.warning(
                "%s: Too many requests for webhook. Blocked for %s seconds. Aborting.",
                webhook,
                ex.retry_after,
            )
            return

        if not response.status_ok:
            webhook.error_queue.enqueue(message)
            logger.warning(
                "%s: Failed to send message to webhook, will retry. "
                "HTTP status code: %d, response: %s",
                webhook,
                response.status_code,
                response.content,
            )

        self.retry(countdown=KILLTRACKER_DISCORD_SEND_DELAY)
    else:
        logger.debug("%s: No more messages to send for webhook", webhook)


@shared_task(timeout=KILLTRACKER_TASKS_TIMEOUT)
def send_test_message_to_webhook(webhook_pk: int, count: int = 1) -> None:
    """send a test message to given webhook.
    Optional inform user about result if user ok is given
    """
    try:
        webhook = Webhook.objects.get(pk=webhook_pk)
    except Webhook.DoesNotExist:
        logger.error("Webhook with pk = %s does not exist", webhook_pk)
        return

    logger.info("Sending %s test messages to webhook %s", count, webhook)
    for n in range(count):
        num_str = f"{n+1}/{count} " if count > 1 else ""
        webhook.enqueue_message(content=f"Test message {num_str}from {__title__}.")

    send_messages_to_webhook.delay(webhook.pk)



from django.contrib import admin

from django.db.models import Q
from django.db.models.functions import Lower
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.utils.safestring import mark_safe

from allianceauth.eveonline.models import EveAllianceInfo
from eveuniverse.models import EveGroup, EveType

from .constants import (
    EVE_CATEGORY_ID_SHIP,
    EVE_CATEGORY_ID_STRUCTURE,
    EVE_CATEGORY_ID_FIGHTER,
    EVE_GROUP_MINING_DRONE,
    EVE_GROUP_ORBITAL_INFRASTRUCTURE,
)
from .core.killmails import Killmail
from .forms import TrackerAdminForm, TrackerAdminKillmailIdForm, field_nice_display
from .models import Webhook, Tracker
from . import tasks


@admin.register(Webhook)
class WebhookAdmin(admin.ModelAdmin):
    list_display = ("name", "is_enabled", "_messages_in_queue")
    list_filter = ("is_enabled",)
    ordering = ("name",)

    def _messages_in_queue(self, obj):
        return obj.main_queue.size()

    actions = ["send_test_message", "purge_messages"]

    def purge_messages(self, request, queryset):
        actions_count = 0
        killmails_deleted = 0
        for webhook in queryset:
            killmails_deleted += webhook.main_queue.clear()
            actions_count += 1

        self.message_user(
            request,
            f"Purged queued messages for {actions_count} webhooks, "
            f"deleting a total of {killmails_deleted} messages.",
        )

    purge_messages.short_description = "Purge queued messages of selected webhooks"

    def send_test_message(self, request, queryset):
        actions_count = 0
        for webhook in queryset:
            tasks.send_test_message_to_webhook.delay(webhook.pk)
            actions_count += 1

        self.message_user(
            request,
            f"Initiated sending of {actions_count} test messages to selected webhooks.",
        )

    send_test_message.short_description = "Send test message to selected webhooks"


@admin.register(Tracker)
class TrackerAdmin(admin.ModelAdmin):
    form = TrackerAdminForm
    list_display = (
        "name",
        "is_enabled",
        "webhook",
        "identify_fleets",
        "_clauses",
        "_pings",
        "_color",
    )
    list_filter = (
        "is_enabled",
        ("origin_solar_system", admin.RelatedOnlyFieldListFilter),
        ("webhook", admin.RelatedOnlyFieldListFilter),
    )
    ordering = ("name",)

    def _color(self, obj):
        html = (
            f'<input type="color" value="{obj.color}" disabled>' if obj.color else "-"
        )
        return mark_safe(html)

    def _pings(self, obj):
        parts = [f"@{group.name}" for group in obj.ping_groups.all()]
        if obj.ping_type != Tracker.ChannelPingType.NONE:
            parts.append(obj.get_ping_type_display())

        return sorted(parts, key=str.casefold) if parts else None

    def _clauses(self, obj):
        clauses = list()
        for field, func in [
            ("origin_solar_system", self._add_to_clauses_1),
            ("require_max_jumps", self._add_to_clauses_1),
            ("require_max_distance", self._add_to_clauses_1),
            ("exclude_attacker_alliances", self._add_to_clauses_2),
            ("exclude_attacker_corporations", self._add_to_clauses_2),
            ("require_attacker_alliances", self._add_to_clauses_2),
            ("require_attacker_corporations", self._add_to_clauses_2),
            ("require_victim_alliances", self._add_to_clauses_2),
            ("require_victim_corporations", self._add_to_clauses_2),
            ("exclude_blue_attackers", self._add_to_clauses_1),
            ("require_blue_victim", self._add_to_clauses_1),
            ("require_min_attackers", self._add_to_clauses_1),
            ("require_max_attackers", self._add_to_clauses_1),
            ("exclude_high_sec", self._add_to_clauses_1),
            ("exclude_low_sec", self._add_to_clauses_1),
            ("exclude_null_sec", self._add_to_clauses_1),
            ("exclude_w_space", self._add_to_clauses_1),
            ("require_regions", self._add_to_clauses_2),
            ("require_constellations", self._add_to_clauses_2),
            ("require_solar_systems", self._add_to_clauses_2),
            ("require_min_value", self._add_to_clauses_1),
            ("require_attackers_ship_groups", self._add_to_clauses_2),
            ("require_attackers_ship_types", self._add_to_clauses_2),
            ("require_victim_ship_groups", self._add_to_clauses_2),
            ("require_victim_ship_types", self._add_to_clauses_2),
            ("exclude_npc_kills", self._add_to_clauses_1),
            ("require_npc_kills", self._add_to_clauses_1),
        ]:
            func(clauses, obj, field)
        return mark_safe("<br>".join(clauses)) if clauses else None

    def _add_to_clauses_1(self, clauses, obj, field_name):
        field = getattr(obj, field_name)
        if field:
            self._append_field_to_clauses(clauses, field_name, getattr(obj, field_name))

    def _add_to_clauses_2(self, clauses, obj, field):
        if getattr(obj, field).count() > 0:
            text = ", ".join(map(str, getattr(obj, field).all()))
            self._append_field_to_clauses(clauses, field, text)

    def _append_field_to_clauses(self, clauses, field, text):
        clauses.append(f"{field_nice_display(field)} = {text}")

    actions = ["disable_tracker", "enable_tracker", "reset_color", "run_test_killmail"]

    def reset_color(self, request, queryset):
        queryset.update(color="")

    reset_color.short_description = "Reset color for selected trackers"

    def enable_tracker(self, request, queryset):
        queryset.update(is_enabled=True)
        self.message_user(request, f"{queryset.count()} trackers enabled.")

    enable_tracker.short_description = "Enable selected trackers"

    def disable_tracker(self, request, queryset):
        queryset.update(is_enabled=False)
        self.message_user(request, f"{queryset.count()} trackers disabled.")

    disable_tracker.short_description = "Disable selected trackers"

    def run_test_killmail(self, request, queryset):
        if "apply" in request.POST:
            form = TrackerAdminKillmailIdForm(request.POST)
            if form.is_valid():
                killmail_id = form.cleaned_data["killmail_id"]
                killmail = Killmail.create_from_zkb_api(killmail_id)
                if killmail:
                    request.session["last_killmail_id"] = killmail_id
                    actions_count = 0
                    for tracker in queryset:
                        tasks.run_tracker.delay(
                            tracker_pk=tracker.pk,
                            killmail_json=killmail.asjson(),
                            ignore_max_age=True,
                        )
                        actions_count += 1

                    self.message_user(
                        request,
                        (
                            f"Started {actions_count} tracker(s) for "
                            f"killmail with ID {killmail_id}."
                        ),
                    )
                else:
                    self.message_user(
                        request,
                        "Failed to load killmail with ID {killmail_id} from ZKB",
                    )

            return HttpResponseRedirect(request.get_full_path())
        else:
            last_killmail_id = request.session.get("last_killmail_id")
            if last_killmail_id:
                initial = {"killmail_id": last_killmail_id}
            else:
                initial = None
            form = TrackerAdminKillmailIdForm(initial=initial)

        return render(
            request,
            "admin/killtracker_killmail_id.html",
            {
                "form": form,
                "title": "Load Test Killmail for Tracker",
                "queryset": queryset.all(),
            },
        )

    run_test_killmail.short_description = "Run test killmail with selected trackers"

    autocomplete_fields = ["origin_solar_system"]

    filter_horizontal = (
        "exclude_attacker_alliances",
        "exclude_attacker_corporations",
        "require_attacker_alliances",
        "require_attacker_corporations",
        "require_victim_alliances",
        "require_victim_corporations",
        "require_regions",
        "require_constellations",
        "require_solar_systems",
        "require_attackers_ship_groups",
        "require_attackers_ship_types",
        "require_victim_ship_groups",
        "require_victim_ship_types",
        "ping_groups",
    )

    fieldsets = (
        (None, {"fields": ("name", "description", "is_enabled", "color")}),
        (
            "Discord Configuration",
            {
                "fields": (
                    "webhook",
                    "ping_type",
                    "ping_groups",
                    "is_posting_name",
                ),
            },
        ),
        (
            "Locations",
            {
                "fields": (
                    "origin_solar_system",
                    "require_max_jumps",
                    "require_max_distance",
                    (
                        "exclude_low_sec",
                        "exclude_null_sec",
                        "exclude_w_space",
                        "exclude_high_sec",
                    ),
                    "require_regions",
                    "require_constellations",
                    "require_solar_systems",
                ),
            },
        ),
        (
            "Organizations",
            {
                "fields": (
                    "exclude_attacker_alliances",
                    "exclude_attacker_corporations",
                    "require_attacker_alliances",
                    "require_attacker_corporations",
                    "require_victim_alliances",
                    "require_victim_corporations",
                ),
            },
        ),
        (
            "Fleet detection",
            {
                "fields": (
                    "require_min_attackers",
                    "require_max_attackers",
                    "identify_fleets",
                ),
            },
        ),
        (
            "EveKillmail properties",
            {
                "fields": (
                    "require_min_value",
                    "exclude_npc_kills",
                    "require_npc_kills",
                ),
            },
        ),
        (
            "Ship types",
            {
                "fields": (
                    "require_attackers_ship_groups",
                    "require_attackers_ship_types",
                    "require_victim_ship_groups",
                    "require_victim_ship_types",
                ),
            },
        ),
    )

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        """overriding this formfield to have sorted lists in the form"""
        if db_field.name == "exclude_attacker_alliances":
            kwargs["queryset"] = EveAllianceInfo.objects.all().order_by(
                Lower("alliance_name")
            )
        elif db_field.name == "require_attacker_alliances":
            kwargs["queryset"] = EveAllianceInfo.objects.all().order_by(
                Lower("alliance_name")
            )
        elif db_field.name == "require_victim_alliances":
            kwargs["queryset"] = EveAllianceInfo.objects.all().order_by(
                Lower("alliance_name")
            )
        elif db_field.name == "require_attackers_ship_groups":
            kwargs["queryset"] = EveGroup.objects.filter(
                eve_category_id__in=[
                    EVE_CATEGORY_ID_STRUCTURE,
                    EVE_CATEGORY_ID_SHIP,
                    EVE_CATEGORY_ID_FIGHTER,
                ],
                published=True,
            ).order_by(Lower("name"))
        elif db_field.name == "require_attackers_ship_types":
            kwargs["queryset"] = (
                EveType.objects.select_related("eve_group")
                .filter(
                    eve_group__eve_category_id__in=[
                        EVE_CATEGORY_ID_STRUCTURE,
                        EVE_CATEGORY_ID_SHIP,
                        EVE_CATEGORY_ID_FIGHTER,
                    ],
                    published=True,
                )
                .order_by(Lower("name"))
            )
        elif db_field.name == "require_victim_ship_groups":
            kwargs["queryset"] = EveGroup.objects.filter(
                (
                    Q(
                        eve_category_id__in=[
                            EVE_CATEGORY_ID_STRUCTURE,
                            EVE_CATEGORY_ID_SHIP,
                            EVE_CATEGORY_ID_FIGHTER,
                        ]
                    )
                    & Q(published=True)
                )
                | (Q(id=EVE_GROUP_MINING_DRONE) & Q(published=True))
                | Q(id=EVE_GROUP_ORBITAL_INFRASTRUCTURE)
            ).order_by(Lower("name"))

        elif db_field.name == "require_victim_ship_types":
            kwargs["queryset"] = (
                EveType.objects.select_related("eve_group")
                .filter(
                    (
                        Q(
                            eve_group__eve_category_id__in=[
                                EVE_CATEGORY_ID_STRUCTURE,
                                EVE_CATEGORY_ID_SHIP,
                                EVE_CATEGORY_ID_FIGHTER,
                            ]
                        )
                        & Q(published=True)
                    )
                    | (Q(eve_group_id=EVE_GROUP_MINING_DRONE) & Q(published=True))
                    | Q(eve_group_id=EVE_GROUP_ORBITAL_INFRASTRUCTURE)
                )
                .order_by(Lower("name"))
            )

        return super().formfield_for_manytomany(db_field, request, **kwargs)



# Generated by Django 3.1.4 on 2021-01-03 23:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("killtracker", "0001_squashed_all"),
    ]

    operations = [
        migrations.AlterField(
            model_name="webhook",
            name="notes",
            field=models.TextField(
                blank=True,
                help_text="you can add notes about this webhook here if you want",
            ),
        ),
    ]



# Generated by Django 2.2.13 on 2020-07-23 22:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveuniverse", "0002_load_eveunit"),
        ("eveonline", "0012_index_additions"),
    ]

    operations = [
        migrations.CreateModel(
            name="EveKillmail",
            fields=[
                ("id", models.BigIntegerField(primary_key=True, serialize=False)),
                (
                    "time",
                    models.DateTimeField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "solar_system",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.EveEntity",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Webhook",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="short name to identify this webhook",
                        max_length=64,
                        unique=True,
                    ),
                ),
                (
                    "webhook_type",
                    models.IntegerField(
                        choices=[(1, "Discord Webhook")],
                        default=1,
                        help_text="type of this webhook",
                    ),
                ),
                (
                    "url",
                    models.CharField(
                        help_text="URL of this webhook, e.g. https://discordapp.com/api/webhooks/123456/abcdef",
                        max_length=255,
                        unique=True,
                    ),
                ),
                (
                    "notes",
                    models.TextField(
                        blank=True,
                        default=None,
                        help_text="you can add notes about this webhook here if you want",
                        null=True,
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        db_index=True,
                        default=True,
                        help_text="whether notifications are currently sent to this webhook",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EveKillmailPosition",
            fields=[
                (
                    "killmail",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="position",
                        serialize=False,
                        to="killtracker.EveKillmail",
                    ),
                ),
                ("x", models.FloatField(blank=True, default=None, null=True)),
                ("y", models.FloatField(blank=True, default=None, null=True)),
                ("z", models.FloatField(blank=True, default=None, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="EveKillmailZkb",
            fields=[
                (
                    "killmail",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="zkb",
                        serialize=False,
                        to="killtracker.EveKillmail",
                    ),
                ),
                (
                    "location_id",
                    models.PositiveIntegerField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                ("hash", models.CharField(blank=True, default="", max_length=64)),
                (
                    "fitted_value",
                    models.FloatField(blank=True, default=None, null=True),
                ),
                (
                    "total_value",
                    models.FloatField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "points",
                    models.PositiveIntegerField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "is_npc",
                    models.BooleanField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "is_solo",
                    models.BooleanField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "is_awox",
                    models.BooleanField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Tracker",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="name to identify tracker. Will be shown on alerts posts.",
                        max_length=100,
                        unique=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        default="",
                        help_text="Brief description what this tracker is for. Will not be shown on alerts.",
                    ),
                ),
                (
                    "require_max_jumps",
                    models.PositiveIntegerField(
                        blank=True,
                        default=None,
                        help_text="Require all killmails to be max x jumps away from origin solar system",
                        null=True,
                    ),
                ),
                (
                    "require_max_distance",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="Require all killmails to be max x LY away from origin solar system",
                        null=True,
                    ),
                ),
                (
                    "identify_fleets",
                    models.BooleanField(
                        default=False,
                        help_text="when true: kills are interpreted and shown as fleet kills",
                    ),
                ),
                (
                    "exclude_blue_attackers",
                    models.BooleanField(
                        default=False, help_text="exclude killmails with blue attackers"
                    ),
                ),
                (
                    "require_blue_victim",
                    models.BooleanField(
                        default=False,
                        help_text="only include killmails where the victim has standing with our group",
                    ),
                ),
                (
                    "require_min_attackers",
                    models.PositiveIntegerField(
                        blank=True,
                        default=None,
                        help_text="Require killmails to have at least given number of attackers",
                        null=True,
                    ),
                ),
                (
                    "require_max_attackers",
                    models.PositiveIntegerField(
                        blank=True,
                        default=None,
                        help_text="Require killmails to have no more than max number of attackers",
                        null=True,
                    ),
                ),
                (
                    "exclude_high_sec",
                    models.BooleanField(
                        default=False,
                        help_text="exclude killmails from high sec. Also exclude high sec systems in route finder for jumps from origin.",
                    ),
                ),
                (
                    "exclude_low_sec",
                    models.BooleanField(
                        default=False, help_text="exclude killmails from low sec"
                    ),
                ),
                (
                    "exclude_null_sec",
                    models.BooleanField(
                        default=False, help_text="exclude killmails from null sec"
                    ),
                ),
                (
                    "exclude_w_space",
                    models.BooleanField(
                        default=False, help_text="exclude killmails from WH space"
                    ),
                ),
                (
                    "require_min_value",
                    models.PositiveIntegerField(
                        blank=True,
                        default=None,
                        help_text="Require killmail's value to be greater or equal to the given value in M ISK",
                        null=True,
                    ),
                ),
                (
                    "exclude_npc_kills",
                    models.BooleanField(default=False, help_text="exclude npc kills"),
                ),
                (
                    "require_npc_kills",
                    models.BooleanField(
                        default=False,
                        help_text="only include killmails that are npc kills",
                    ),
                ),
                (
                    "ping_type",
                    models.CharField(
                        choices=[
                            ("PN", "(no ping)"),
                            ("PH", "@here"),
                            ("PE", "@everybody"),
                        ],
                        default="PN",
                        help_text="Options for pinging on every matching killmail",
                        max_length=2,
                    ),
                ),
                (
                    "is_posting_name",
                    models.BooleanField(
                        default=True,
                        help_text="whether posted messages include the tracker's name",
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        db_index=True,
                        default=True,
                        help_text="toogle for activating or deactivating a tracker",
                    ),
                ),
                (
                    "exclude_attacker_alliances",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="exclude killmails with attackers from one of these alliances",
                        related_name="tracker_exclude_attacker_alliances_set",
                        to="eveonline.EveAllianceInfo",
                    ),
                ),
                (
                    "exclude_attacker_corporations",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="exclude killmails with attackers from one of these corporations",
                        related_name="tracker_exclude_attacker_corporations_set",
                        to="eveonline.EveCorporationInfo",
                    ),
                ),
                (
                    "origin_solar_system",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="Solar system to calculate distance and jumps from. When provided distance and jumps will be shown on killmail messages",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="tracker_origin_solar_systems_set",
                        to="eveuniverse.EveSolarSystem",
                    ),
                ),
                (
                    "require_attacker_alliances",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="only include killmails with attackers from one of these alliances",
                        related_name="tracker_required_attacker_alliances_set",
                        to="eveonline.EveAllianceInfo",
                    ),
                ),
                (
                    "require_attacker_corporations",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="only include killmails with attackers from one of these corporations",
                        related_name="tracker_required_attacker_corporations_set",
                        to="eveonline.EveCorporationInfo",
                    ),
                ),
                (
                    "require_attackers_ship_groups",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails where at least one attacker is flying one of these ship groups",
                        limit_choices_to=models.Q(
                            models.Q(
                                ("eve_category_id", 65),
                                ("eve_category_id", 6),
                                _connector="OR",
                            ),
                            ("published", True),
                        ),
                        related_name="tracker_require_attackers_ship_groups_set",
                        to="eveuniverse.EveGroup",
                    ),
                ),
                (
                    "require_attackers_ship_types",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails where at least one attacker is flying one of these ship types",
                        limit_choices_to=models.Q(
                            models.Q(
                                ("eve_group__eve_category_id", 65),
                                ("eve_group__eve_category_id", 6),
                                _connector="OR",
                            ),
                            ("published", True),
                        ),
                        related_name="tracker_require_attackers_ship_groups_set",
                        to="eveuniverse.EveType",
                    ),
                ),
                (
                    "require_constellations",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails that occurred in one of these regions",
                        to="eveuniverse.EveConstellation",
                    ),
                ),
                (
                    "require_regions",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails that occurred in one of these regions",
                        to="eveuniverse.EveRegion",
                    ),
                ),
                (
                    "require_solar_systems",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails that occurred in one of these regions",
                        related_name="tracker_require_solar_systems_set",
                        to="eveuniverse.EveSolarSystem",
                    ),
                ),
                (
                    "require_victim_alliances",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="only include killmails where the victim belongs to one of these alliances",
                        related_name="tracker_require_victim_alliances_set",
                        to="eveonline.EveAllianceInfo",
                    ),
                ),
                (
                    "require_victim_corporations",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="only include killmails where the victim belongs to one of these corporations",
                        related_name="tracker_require_victim_corporations_set",
                        to="eveonline.EveCorporationInfo",
                    ),
                ),
                (
                    "require_victim_ship_groups",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails where victim is flying one of these ship groups",
                        limit_choices_to=models.Q(
                            models.Q(
                                ("eve_category_id", 65),
                                ("eve_category_id", 6),
                                _connector="OR",
                            ),
                            ("published", True),
                        ),
                        related_name="tracker_require_victim_ship_groups_set",
                        to="eveuniverse.EveGroup",
                    ),
                ),
                (
                    "webhook",
                    models.ForeignKey(
                        help_text="Webhook URL for a channel on Discord to sent all alerts to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="killtracker.Webhook",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EveKillmailAttacker",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "damage_done",
                    models.BigIntegerField(blank=True, default=None, null=True),
                ),
                (
                    "is_final_blow",
                    models.BooleanField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "security_status",
                    models.FloatField(blank=True, default=None, null=True),
                ),
                (
                    "alliance",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="evekillmailattacker_alliances_set",
                        to="eveuniverse.EveEntity",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="evekillmailattacker_characters_set",
                        to="eveuniverse.EveEntity",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="evekillmailattacker_corporations_set",
                        to="eveuniverse.EveEntity",
                    ),
                ),
                (
                    "faction",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="evekillmailattacker_factions_set",
                        to="eveuniverse.EveEntity",
                    ),
                ),
                (
                    "killmail",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="attackers",
                        to="killtracker.EveKillmail",
                    ),
                ),
                (
                    "ship_type",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="evekillmailattacker_shiptypes_set",
                        to="eveuniverse.EveEntity",
                    ),
                ),
                (
                    "weapon_type",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.EveEntity",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="EveKillmailVictim",
            fields=[
                (
                    "killmail",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="victim",
                        serialize=False,
                        to="killtracker.EveKillmail",
                    ),
                ),
                (
                    "damage_taken",
                    models.BigIntegerField(blank=True, default=None, null=True),
                ),
                (
                    "alliance",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="evekillmailvictim_alliances_set",
                        to="eveuniverse.EveEntity",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="evekillmailvictim_characters_set",
                        to="eveuniverse.EveEntity",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="evekillmailvictim_corporations_set",
                        to="eveuniverse.EveEntity",
                    ),
                ),
                (
                    "faction",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="evekillmailvictim_factions_set",
                        to="eveuniverse.EveEntity",
                    ),
                ),
                (
                    "ship_type",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="evekillmailvictim_shiptypes_set",
                        to="eveuniverse.EveEntity",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



# flake8: noqa
""" this generated the list of current tracker conditions for the documentation"""

# init and setup django project
import inspect
import os
import sys

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
myauth_dir = os.path.dirname(os.path.dirname(os.path.dirname(currentdir))) + "/myauth"
sys.path.insert(0, myauth_dir)

import django

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

# normal includes
from killtracker.models import Tracker

my_fields = [
    f
    for f in Tracker._meta.get_fields()
    if f.name.split("_")[0] in ["exclude", "require"]
]

for field in my_fields:
    field_name = field.name.replace("_", " ")
    print(f"{field_name}|{field.help_text}")



# Generated by Django 3.1.4 on 2021-01-03 23:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("eveuniverse", "0004_effect_longer_name"),
        ("eveonline", "0012_index_additions"),
        ("killtracker", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="tracker",
            name="color",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Optional color for embed on Discord - #000000 / black means no color selected",
                max_length=7,
            ),
        ),
        migrations.AddField(
            model_name="tracker",
            name="ping_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Option to ping specific group members - ",
                related_name="_tracker_ping_groups_+",
                to="auth.Group",
                verbose_name="group pings",
            ),
        ),
        migrations.AddField(
            model_name="tracker",
            name="require_victim_ship_types",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails where victim is flying one of these ship types",
                related_name="_tracker_require_victim_ship_types_+",
                to="eveuniverse.EveType",
            ),
        ),
        migrations.AlterField(
            model_name="evekillmailattacker",
            name="alliance",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evekillmailattacker",
            name="character",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evekillmailattacker",
            name="corporation",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evekillmailattacker",
            name="faction",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evekillmailattacker",
            name="ship_type",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evekillmailattacker",
            name="weapon_type",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evekillmailvictim",
            name="alliance",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evekillmailvictim",
            name="character",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evekillmailvictim",
            name="corporation",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evekillmailvictim",
            name="faction",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="evekillmailvictim",
            name="ship_type",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="description",
            field=models.TextField(
                blank=True,
                help_text="Brief description what this tracker is for. Will not be shown on alerts.",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="exclude_attacker_alliances",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="exclude killmails with attackers from one of these alliances",
                related_name="_tracker_exclude_attacker_alliances_+",
                to="eveonline.EveAllianceInfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="exclude_attacker_corporations",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="exclude killmails with attackers from one of these corporations",
                related_name="_tracker_exclude_attacker_corporations_+",
                to="eveonline.EveCorporationInfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="origin_solar_system",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="Solar system to calculate distance and jumps from. When provided distance and jumps will be shown on killmail messages",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="eveuniverse.evesolarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="ping_type",
            field=models.CharField(
                choices=[("PN", "(none)"), ("PH", "@here"), ("PE", "@everybody")],
                default="PN",
                help_text="Option to ping every member of the channel",
                max_length=2,
                verbose_name="channel pings",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_attacker_alliances",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="only include killmails with attackers from one of these alliances",
                related_name="_tracker_require_attacker_alliances_+",
                to="eveonline.EveAllianceInfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_attacker_corporations",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="only include killmails with attackers from one of these corporations",
                related_name="_tracker_require_attacker_corporations_+",
                to="eveonline.EveCorporationInfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_attackers_ship_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails where at least one attacker is flying one of these ship groups",
                related_name="_tracker_require_attackers_ship_groups_+",
                to="eveuniverse.EveGroup",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_attackers_ship_types",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails where at least one attacker is flying one of these ship types",
                related_name="_tracker_require_attackers_ship_types_+",
                to="eveuniverse.EveType",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_constellations",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails that occurred in one of these regions",
                related_name="_tracker_require_constellations_+",
                to="eveuniverse.EveConstellation",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_regions",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails that occurred in one of these regions",
                related_name="_tracker_require_regions_+",
                to="eveuniverse.EveRegion",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_solar_systems",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails that occurred in one of these regions",
                related_name="_tracker_require_solar_systems_+",
                to="eveuniverse.EveSolarSystem",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_victim_alliances",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="only include killmails where the victim belongs to one of these alliances",
                related_name="_tracker_require_victim_alliances_+",
                to="eveonline.EveAllianceInfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_victim_corporations",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="only include killmails where the victim belongs to one of these corporations",
                related_name="_tracker_require_victim_corporations_+",
                to="eveonline.EveCorporationInfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_victim_ship_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails where victim is flying one of these ship groups",
                related_name="_tracker_require_victim_ship_groups_+",
                to="eveuniverse.EveGroup",
            ),
        ),
        migrations.AlterField(
            model_name="webhook",
            name="notes",
            field=models.TextField(
                blank=True,
                help_text="you can add notes about this webhook here if you want",
            ),
            preserve_default=False,
        ),
    ]



from datetime import timedelta

from typing import Dict, Tuple

from django.db import models, transaction
from django.utils.timezone import now

from allianceauth.services.hooks import get_extension_logger

from app_utils.caching import ObjectCacheMixin
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveEntity

from . import __title__
from .app_settings import KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS
from .core.killmails import Killmail, _KillmailCharacter

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class EveKillmailQuerySet(models.QuerySet):
    """Custom queryset for EveKillmail"""

    def load_entities(self) -> int:
        """loads unknown entities for all killmails of this QuerySet.
        Returns count of updated entities
        """
        entity_ids = []
        for killmail in self:
            entity_ids += killmail.entity_ids()

        return EveEntity.objects.filter(
            id__in=list(set(entity_ids)), name=""
        ).update_from_esi()


class EveKillmailManager(models.Manager):
    def get_queryset(self) -> models.QuerySet:
        return EveKillmailQuerySet(self.model, using=self._db)

    def delete_stale(self) -> Tuple[int, Dict[str, int]]:
        """deletes all stale killmail"""
        if KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS > 0:
            deadline = now() - timedelta(days=KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS)
            return self.filter(time__lt=deadline).delete()

    def create_from_killmail(
        self, killmail: Killmail, resolve_ids=True
    ) -> models.Model:
        """create a new EveKillmail from a Killmail object and returns it

        Args:
        - resolve_ids: When set to False will not resolve EveEntity IDs

        """
        from .models import (
            EveKillmailAttacker,
            EveKillmailPosition,
            EveKillmailVictim,
            EveKillmailZkb,
        )

        args = {
            "id": killmail.id,
            "time": killmail.time,
        }
        if killmail.solar_system_id:
            args["solar_system"], _ = EveEntity.objects.get_or_create(
                id=killmail.solar_system_id
            )
        eve_killmail = self.create(**args)

        if killmail.zkb:
            args = {**killmail.zkb.asdict(), **{"killmail": eve_killmail}}
            EveKillmailZkb.objects.create(**args)

        args = {
            **{
                "killmail": eve_killmail,
                "damage_taken": killmail.victim.damage_taken,
            },
            **self._create_args_for_entities(killmail.victim),
        }
        EveKillmailVictim.objects.create(**args)

        args = {**killmail.position.asdict(), **{"killmail": eve_killmail}}
        EveKillmailPosition.objects.create(**args)

        for attacker in killmail.attackers:
            args = {
                **{
                    "killmail": eve_killmail,
                    "damage_done": attacker.damage_done,
                    "security_status": attacker.security_status,
                    "is_final_blow": attacker.is_final_blow,
                },
                **self._create_args_for_entities(attacker),
            }
            EveKillmailAttacker.objects.create(**args)

        if resolve_ids:
            EveEntity.objects.bulk_update_new_esi()

        eve_killmail.refresh_from_db()
        return eve_killmail

    @staticmethod
    def _create_args_for_entities(killmail_character: _KillmailCharacter) -> dict:
        args = dict()
        for prop_name in killmail_character.ENTITY_PROPS:
            entity_id = getattr(killmail_character, prop_name)
            if entity_id:
                field = prop_name.replace("_id", "")
                args[field], _ = EveEntity.objects.get_or_create(id=entity_id)

        return args

    def update_or_create_from_killmail(
        self, killmail: Killmail
    ) -> Tuple[models.Model, bool]:
        with transaction.atomic():
            try:
                self.get(id=killmail.id).delete()
                created = False
            except self.model.DoesNotExist:
                created = True

            obj = self.create_from_killmail(killmail, resolve_ids=False)

        if created:
            EveEntity.objects.bulk_update_new_esi()

        return obj, created


class TrackerManager(ObjectCacheMixin, models.Manager):
    pass


class WebhookManager(ObjectCacheMixin, models.Manager):
    pass



from app_utils.django import clean_setting


# ignore killmails that are older than the given number in minutes
# sometimes killmails appear belated on ZKB,
# this feature ensures they don't create new alerts
KILLTRACKER_KILLMAIL_MAX_AGE_FOR_TRACKER = clean_setting(
    "KILLTRACKER_KILLMAIL_MAX_AGE_FOR_TRACKER", 60
)

# Maximum number of killmails retrieved from ZKB by task run
KILLTRACKER_MAX_KILLMAILS_PER_RUN = clean_setting(
    "KILLTRACKER_MAX_KILLMAILS_PER_RUN", 200
)

# Killmails older than set number of days will be purged from the database.
# If you want to keep all killmails set this to 0.
KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS = clean_setting(
    "KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS", 30
)

# whether killmails retrieved from ZKB are stored in the database
KILLTRACKER_STORING_KILLMAILS_ENABLED = clean_setting(
    "KILLTRACKER_STORING_KILLMAILS_ENABLED", False
)

# Wether app sets the name and avatar icon of a webhook.
# When False the webhook will use it's own values as set on the platform
KILLTRACKER_WEBHOOK_SET_AVATAR = clean_setting("KILLTRACKER_WEBHOOK_SET_AVATAR", True)


#####################
# INTERNAL SETTINGS

# Max duration to wait for new killmails from redisq in seconds
KILLTRACKER_REDISQ_TTW = clean_setting("KILLTRACKER_REDISQ_TTW", 5)

# Tasks hard timeout
KILLTRACKER_TASKS_TIMEOUT = clean_setting("KILLTRACKER_TASKS_TIMEOUT", 1800)

# delay in seconds between every message sent to Discord
# this needs to be >= 1 to prevent 429 Too Many Request errors
KILLTRACKER_DISCORD_SEND_DELAY = clean_setting(
    "KILLTRACKER_DISCORD_SEND_DELAY", default_value=2, min_value=1, max_value=900
)

# Maximum retries when generating a message from a killmail
KILLTRACKER_GENERATE_MESSAGE_MAX_RETRIES = clean_setting(
    "KILLTRACKER_GENERATE_MESSAGE_MAX_RETRIES", 3
)

# Delay when retrying to generate a message in seconds
KILLTRACKER_GENERATE_MESSAGE_RETRY_COUNTDOWN = clean_setting(
    "KILLTRACKER_GENERATE_MESSAGE_RETRY_COUNTDOWN", 10
)

# Cache duration for objects in tasks in seconds
KILLTRACKER_TASK_OBJECTS_CACHE_TIMEOUT = clean_setting(
    "KILLTRACKER_TASK_OBJECTS_CACHE_TIMEOUT", 60
)

# Minimum delay when retrying a task
KILLTRACKER_TASK_MINIMUM_RETRY_DELAY = clean_setting(
    "KILLTRACKER_TASK_MINIMUM_RETRY_DELAY", default_value=0.05, min_value=0.0
)



default_app_config = "killtracker.apps.KillmailsConfig"

__version__ = "0.3.0"
__title__ = "Killtracker"

APP_NAME = "aa-killtracker"
HOMEPAGE_URL = "https://gitlab.com/ErikKalkoken/aa-killtracker"
USER_AGENT_TEXT = f"{APP_NAME} v{__version__}"



from django import forms
from django.core.exceptions import ValidationError
from django.forms.widgets import TextInput
from django.utils.translation import gettext_lazy as _

from .models import Tracker


def field_nice_display(name: str) -> str:
    return name.replace("_", " ").capitalize()


class TrackerAdminKillmailIdForm(forms.Form):
    killmail_id = forms.IntegerField()


class TrackerAdminForm(forms.ModelForm):
    class Meta:
        model = Tracker
        fields = "__all__"
        widgets = {
            "color": TextInput(attrs={"type": "color"}),
        }

    def clean(self):
        cleaned_data = super().clean()

        if (
            cleaned_data["require_max_jumps"]
            and not cleaned_data["origin_solar_system"]
        ):
            raise ValidationError(
                {
                    "origin_solar_system": _(
                        "'Require max jumps' needs an "
                        f"{field_nice_display('origin_solar_system')} to work"
                    )
                }
            )

        if (
            cleaned_data["require_max_distance"]
            and not cleaned_data["origin_solar_system"]
        ):
            raise ValidationError(
                {
                    "origin_solar_system": _(
                        "'Require max distance' needs an "
                        f"{field_nice_display('origin_solar_system')} to work"
                    )
                }
            )

        self._validate_not_same_options_chosen(
            cleaned_data,
            "exclude_attacker_alliances",
            "require_attacker_alliances",
        )
        self._validate_not_same_options_chosen(
            cleaned_data,
            "exclude_attacker_corporations",
            "require_attacker_corporations",
        )

        if (
            cleaned_data["require_min_attackers"]
            and cleaned_data["require_max_attackers"]
            and cleaned_data["require_min_attackers"]
            > cleaned_data["require_max_attackers"]
        ):
            raise ValidationError(
                {
                    "require_min_attackers": _(
                        "Can not be larger than "
                        f"{field_nice_display('require_max_attackers')}"
                    )
                }
            )

        if (
            cleaned_data["exclude_high_sec"]
            and cleaned_data["exclude_low_sec"]
            and cleaned_data["exclude_null_sec"]
            and cleaned_data["exclude_w_space"]
        ):
            text = ", ".join(
                [
                    field_nice_display(x)
                    for x in [
                        "exclude_low_sec",
                        "exclude_null_sec",
                        "exclude_w_space",
                        "exclude_high_sec",
                    ]
                ]
            )
            raise ValidationError(
                f"Setting all four clauses together does not make sense: {text}"
            )

        if cleaned_data["exclude_npc_kills"] and cleaned_data["require_npc_kills"]:
            text = ", ".join(
                [
                    field_nice_display(x)
                    for x in [
                        "exclude_npc_kills",
                        "require_npc_kills",
                    ]
                ]
            )
            raise ValidationError(
                f"Setting both clauses together does not make sense: {text}"
            )

    @staticmethod
    def _validate_not_same_options_chosen(
        cleaned_data, field_name_1, field_name_2, display_func=lambda x: x
    ) -> None:
        same_options = set(cleaned_data[field_name_1]).intersection(
            set(cleaned_data[field_name_2])
        )
        if same_options:
            same_options_text = ", ".join(
                map(
                    str,
                    [display_func(x) for x in same_options],
                )
            )
            raise ValidationError(
                f"Can not choose same options for {field_nice_display(field_name_1)} "
                f"& {field_nice_display(field_name_2)}: {same_options_text}"
            )
